/*
 * bsc1176931_core
 *
 * Fix for CVE-2020-0429, bsc#1176931 (l2tp_core.c part)
 *
 *  Copyright (c) 2020 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if !IS_MODULE(CONFIG_L2TP)
#error "Live patch supports only CONFIG_L2TP=m"
#endif

#define KGR_PATCHED_MODULE "l2tp_core"

/* klp-ccp: from net/l2tp/l2tp_core.c */
#define pr_fmt(fmt) KGR_PATCHED_MODULE ": " fmt

#include <linux/module.h>
#include <linux/string.h>
#include <linux/list.h>
#include <linux/rculist.h>
#include <linux/uaccess.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/jiffies.h>
#include <linux/netdevice.h>
#include <linux/net.h>
#include <linux/inetdevice.h>
#include <linux/skbuff.h>
#include <linux/init.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/l2tp.h>
#include <linux/hash.h>
#include <linux/nsproxy.h>
#include <net/net_namespace.h>
#include <net/netns/generic.h>
#include <net/dst.h>
#include <net/inet_ecn.h>
#include <net/ip6_route.h>
#include <asm/byteorder.h>
#include <linux/atomic.h>

#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1176931.h"
#include "../kallsyms_relocs.h"

#include "bsc1176931_common.h"

/* klp-ccp: from net/l2tp/l2tp_core.h */
static void (*kgre___l2tp_session_unhash)(struct l2tp_session *session);

static int (*kgre_l2tp_session_queue_purge)(struct l2tp_session *session);

static void (*kgre_l2tp_session_set_header_len)(struct l2tp_session *session, int version);

static void (*kgre_l2tp_session_free)(struct l2tp_session *session);

static inline void kgrr_l2tp_session_dec_refcount_1(struct l2tp_session *session)
{
	if (atomic_dec_and_test(&session->ref_count))
		(*kgre_l2tp_session_free)(session);
}

#define kgrr_l2tp_session_dec_refcount(s) kgrr_l2tp_session_dec_refcount_1(s)

/* klp-ccp: from net/l2tp/l2tp_core.c */
#define L2TP_HDR_VER_2	   0x0002
#define L2TP_HDR_VER_3	   0x0003

static atomic_t (*kgre_l2tp_session_count);

static unsigned int (*kgre_l2tp_net_id);
struct l2tp_net {
	struct list_head l2tp_tunnel_list;
	spinlock_t l2tp_tunnel_list_lock;
	struct hlist_head l2tp_session_hlist[L2TP_HASH_SIZE_2];
	spinlock_t l2tp_session_hlist_lock;
};

static inline struct l2tp_net *kgrr_l2tp_pernet(struct net *net)
{
	BUG_ON(!net);

	return net_generic(net, (*kgre_l2tp_net_id));
}

static inline void l2tp_tunnel_inc_refcount_1(struct l2tp_tunnel *tunnel)
{
	atomic_inc(&tunnel->ref_count);
}

#define l2tp_tunnel_inc_refcount(t) l2tp_tunnel_inc_refcount_1(t)

static inline struct hlist_head *
l2tp_session_id_hash_2(struct l2tp_net *pn, u32 session_id)
{
	return &pn->l2tp_session_hlist[hash_32(session_id, L2TP_HASH_BITS_2)];

}

static inline struct hlist_head *
l2tp_session_id_hash(struct l2tp_tunnel *tunnel, u32 session_id)
{
	return &tunnel->session_hlist[hash_32(session_id, L2TP_HASH_BITS)];
}

/* New */
static int kgrp_l2tp_session_add_to_tunnel(struct l2tp_tunnel *tunnel,
				     struct l2tp_session *session)
{
	struct l2tp_session *session_walk;
	struct hlist_head *g_head;
	struct hlist_head *head;
	struct l2tp_net *pn;

	head = l2tp_session_id_hash(tunnel, session->session_id);

	write_lock_bh(&tunnel->hlist_lock);
	hlist_for_each_entry(session_walk, head, hlist)
		if (session_walk->session_id == session->session_id)
			goto exist;

	if (tunnel->version == L2TP_HDR_VER_3) {
		pn = kgrr_l2tp_pernet(tunnel->l2tp_net);
		g_head = l2tp_session_id_hash_2(kgrr_l2tp_pernet(tunnel->l2tp_net),
						session->session_id);

		spin_lock_bh(&pn->l2tp_session_hlist_lock);
		hlist_for_each_entry(session_walk, g_head, global_hlist)
			if (session_walk->session_id == session->session_id)
				goto exist_glob;

		hlist_add_head_rcu(&session->global_hlist, g_head);
		spin_unlock_bh(&pn->l2tp_session_hlist_lock);
	}

	hlist_add_head(&session->hlist, head);
	write_unlock_bh(&tunnel->hlist_lock);

	return 0;

exist_glob:
	spin_unlock_bh(&pn->l2tp_session_hlist_lock);
exist:
	write_unlock_bh(&tunnel->hlist_lock);

	return -EEXIST;
}

static int (*kgre_l2tp_build_l2tpv2_header)(struct l2tp_session *session, void *buf);

static int (*kgre_l2tp_build_l2tpv3_header)(struct l2tp_session *session, void *buf);

/* New. Non-zero return: session already marked dead. */
static bool kgrp_mark_session_dead(struct l2tp_session *session)
{
	/*
	 * Abuse the tail of session->name[32] for storing the new
	 * ->dead. ->name[] gets initialized in l2tp_session_create()
	 * by means of sprintf(..., "sess %u/%u", tunnel_id, session_id).
	 * tunnel_id and session_id are of type u32 each and so there
	 * decimal representation is always <= 10 characters in length.
	 * Thus, no more than 27 characters of ->name will ever be
	 * used, including the terminating '\0'. Finally note that
	 * struct l2tp_session instances always get allocated by
	 * means of kzalloc(), so the trailing part of ->name[] will
	 * be initialized to zero.
	 */
	u32 *session_dead;
	u32 oldval, newval;

	session_dead = (int *)&session->name[28];
	newval = 0x00100000;
	oldval = cmpxchg(session_dead, 0, newval);
	if (!oldval)
		return false;

	/*
	 * oldval should equal newval, which indicates that this
	 * session had been marked dead before. Return true in this
	 * case. OTOH, if oldval is neither 0 nor newval, there's
	 * something off; warn and return false to force the caller
	 * into the original, non-patched behaviour.
	 */
	return !WARN_ON_ONCE(oldval != newval);
}

void kgrp_l2tp_tunnel_closeall(struct l2tp_tunnel *tunnel)
{
	int hash;
	struct hlist_node *walk;
	struct hlist_node *tmp;
	struct l2tp_session *session;

	BUG_ON(tunnel == NULL);

	l2tp_info(tunnel, L2TP_MSG_CONTROL, "%s: closing all sessions...\n",
		  tunnel->name);

	write_lock_bh(&tunnel->hlist_lock);
	for (hash = 0; hash < L2TP_HASH_SIZE; hash++) {
again:
		hlist_for_each_safe(walk, tmp, &tunnel->session_hlist[hash]) {
			session = hlist_entry(walk, struct l2tp_session, hlist);

			l2tp_info(session, L2TP_MSG_CONTROL,
				  "%s: closing session\n", session->name);

			hlist_del_init(&session->hlist);

			/*
			 * Fix CVE-2020-0429
			 *  +2 lines
			 */
			if (kgrp_mark_session_dead(session))
				goto again;

			if (session->ref != NULL)
				(*session->ref)(session);

			write_unlock_bh(&tunnel->hlist_lock);

			(*kgre___l2tp_session_unhash)(session);
			(*kgre_l2tp_session_queue_purge)(session);

			if (session->session_close != NULL)
				(*session->session_close)(session);

			if (session->deref != NULL)
				(*session->deref)(session);

			kgrr_l2tp_session_dec_refcount(session);

			write_lock_bh(&tunnel->hlist_lock);

			/* Now restart from the beginning of this hash
			 * chain.  We always remove a session from the
			 * list so we are guaranteed to make forward
			 * progress.
			 */
			goto again;
		}
	}
	write_unlock_bh(&tunnel->hlist_lock);
}

int kgrp_l2tp_session_delete(struct l2tp_session *session)
{
	/*
	 * Fix CVE-2020-0429
	 *  +2 lines
	 */
	if (kgrp_mark_session_dead(session))
		return 0;

	if (session->ref)
		(*session->ref)(session);
	(*kgre___l2tp_session_unhash)(session);
	(*kgre_l2tp_session_queue_purge)(session);
	if (session->session_close != NULL)
		(*session->session_close)(session);
	if (session->deref)
		(*session->deref)(session);
	kgrr_l2tp_session_dec_refcount(session);
	return 0;
}

struct l2tp_session *kgrp_l2tp_session_create(int priv_size, struct l2tp_tunnel *tunnel, u32 session_id, u32 peer_session_id, struct l2tp_session_cfg *cfg)
{
	struct l2tp_session *session;
	/*
	 * Fix CVE-2020-0429
	 *  +1 line
	 */
	int err;

	session = kzalloc(sizeof(struct l2tp_session) + priv_size, GFP_KERNEL);
	if (session != NULL) {
		session->magic = L2TP_SESSION_MAGIC;
		session->tunnel = tunnel;

		session->session_id = session_id;
		session->peer_session_id = peer_session_id;
		session->nr = 0;
		if (tunnel->version == L2TP_HDR_VER_2)
			session->nr_max = 0xffff;
		else
			session->nr_max = 0xffffff;
		session->nr_window_size = session->nr_max / 2;
		session->nr_oos_count_max = 4;

		/* Use NR of first received packet */
		session->reorder_skip = 1;

		sprintf(&session->name[0], "sess %u/%u",
			tunnel->tunnel_id, session->session_id);

		skb_queue_head_init(&session->reorder_q);

		INIT_HLIST_NODE(&session->hlist);
		INIT_HLIST_NODE(&session->global_hlist);

		/* Inherit debug options from tunnel */
		session->debug = tunnel->debug;

		if (cfg) {
			session->pwtype = cfg->pw_type;
			session->debug = cfg->debug;
			session->mtu = cfg->mtu;
			session->mru = cfg->mru;
			session->send_seq = cfg->send_seq;
			session->recv_seq = cfg->recv_seq;
			session->lns_mode = cfg->lns_mode;
			session->reorder_timeout = cfg->reorder_timeout;
			session->offset = cfg->offset;
			session->l2specific_type = cfg->l2specific_type;
			session->l2specific_len = cfg->l2specific_len;
			session->cookie_len = cfg->cookie_len;
			memcpy(&session->cookie[0], &cfg->cookie[0], cfg->cookie_len);
			session->peer_cookie_len = cfg->peer_cookie_len;
			memcpy(&session->peer_cookie[0], &cfg->peer_cookie[0], cfg->peer_cookie_len);
		}

		if (tunnel->version == L2TP_HDR_VER_2)
			session->build_header = (*kgre_l2tp_build_l2tpv2_header);
		else
			session->build_header = (*kgre_l2tp_build_l2tpv3_header);

		(*kgre_l2tp_session_set_header_len)(session, tunnel->version);

		/*
		 * Fix CVE-2020-0429
		 *  +6 lines
		 */
		err = kgrp_l2tp_session_add_to_tunnel(tunnel, session);
		if (err) {
			kfree(session);

			return ERR_PTR(err);
		}

		/* Bump the reference count. The session context is deleted
		 * only when this drops to zero.
		 */
		l2tp_session_inc_refcount(session);
		l2tp_tunnel_inc_refcount(tunnel);

		/* Ensure tunnel socket isn't deleted */
		sock_hold(tunnel->sock);

		/*
		 * Fix CVE-2020-0429
		 *  -15 lines
		 */

		/* Ignore management session in session count value */
		if (session->session_id != 0)
			atomic_inc(&(*kgre_l2tp_session_count));

		/*
		 * Fix CVE-2020-0429
		 *  +1 line
		 */
		return session;
	}

	/*
	 * Fix CVE-2020-0429
	 *  -1 line, +1 line
	 */
	return ERR_PTR(-ENOMEM);
}



static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "l2tp_net_id", (void *)&kgre_l2tp_net_id, "l2tp_core" },
	{ "l2tp_session_count", (void *)&kgre_l2tp_session_count, "l2tp_core" },
	{ "__l2tp_session_unhash", (void *)&kgre___l2tp_session_unhash,
	  "l2tp_core" },
	{ "l2tp_session_queue_purge", (void *)&kgre_l2tp_session_queue_purge,
	  "l2tp_core" },
	{ "l2tp_session_set_header_len",
	  (void *)&kgre_l2tp_session_set_header_len, "l2tp_core" },
	{ "l2tp_session_free", (void *)&kgre_l2tp_session_free, "l2tp_core" },
	{ "l2tp_build_l2tpv2_header", (void *)&kgre_l2tp_build_l2tpv2_header,
	  "l2tp_core" },
	{ "l2tp_build_l2tpv3_header", (void *)&kgre_l2tp_build_l2tpv3_header,
	  "l2tp_core" },
};

static int kgr_patch_bsc1176931_core_module_notify(struct notifier_block *nb,
						   unsigned long action,
						   void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1176931_core_module_nb = {
	.notifier_call = kgr_patch_bsc1176931_core_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1176931_core_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1176931_core_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1176931_core_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1176931_core_module_nb);
}
